/*
  LOS.C
  ERKKI-enginen LOS(Line Of Sight)-rutiinit.

  Versio 0.8.5

  Tarmo Korpela 2005

  All of these prosedures use the engines field structure, KENTTA,
  as a parameter. If you're not using engines field structure, you
  can make a clone structure including only the data-member of the
  structure and typecasting it as a KENTTA (defined in ERKKI.H).

  Sama suomeksi: kaikki LOS-rutiinit kyttvt enginen KENTTA-
  rakennetta. Jos et kyt enginest muuta, voit mritell
  oman KENTTA-kloonirakenteen (eli muokkaa erkki.h uuteen uskoon)

*/

/* Include the headers needed - LOS.H 'n' ERKKI.H - Tarvittavat otsikkotiedostot*/
#include "LOS.H"

/*
  Is the location (x,y) exposed? If so, returns ERKKI_LOS_SEEN, else zero.
  Onko sijainti (x,y) paljastettu? Jos, palauttaa ERKKI_LOS_SEEN, muuten 0.

  Esim. if( ERKKI_isexposed(field, 12, 34) != 0)
          printf("Sijainti (12, 34) on nkyviss!");

*/
int ERKKI_isexposed(KENTTA * kentta, signed int x, signed int y){
  if(NOTINMAP(kentta, x, y))return 0; /* We're out from field! */
  return (kentta->Kentta[x+y*kentta->w].data & ERKKI_LOS_SEEN);
}

/*
   Is the location (x,y) visible (i.e. are the light power flags on?)
   Returns the light power of the tile (0-255), 0 meanining non-visibility.
*/
int ERKKI_isvisible(KENTTA * kentta, signed int x, signed int y){
  if(NOTINMAP(kentta, x, y))return 0; /* Again out! */
  return (kentta->Kentta[x+y*kentta->w].data & ERKKI_LOS_LIGHT);
}

/*
   Removes maps light, i.e. sets all light power flags to zero.
*/
void ERKKI_unsee_area(KENTTA * kentta){
int x, y, k;
  for(x=0;x<kentta->w;x++)for(y=0;y<kentta->h;y++){
    k = x+y*kentta->w; /* Location in table */
    /* Does the tile have lights on? If, *click*. */
    if(kentta->Kentta[k].data & ERKKI_LOS_LIGHT){
      kentta->Kentta[k].data ^= (kentta->Kentta[k].data & ERKKI_LOS_LIGHT);
    }
  }
}

/*
  Spots the light in location (x,y) (if possible) with light power teho.
*/
void ERKKI_spot_light(KENTTA * kentta, signed int x, signed int y, unsigned char teho){
signed int k;
signed int bb;
  if( NOTINMAP(kentta, x, y) )return;
  k = x + (y * kentta->w); /* Location in structures table */
  bb = kentta->Kentta[k].data;
  if((bb & ERKKI_LOS_SEEN) == 0){ /* This block isn't even exposed! */
    bb |= ERKKI_LOS_SEEN; /* Now it's exposed */
    /* Change the light power, if it's greater than original.
       This looks weird, but I'm only zeroing light without touching other bytes. */
    if((bb & ERKKI_LOS_LIGHT) < teho)bb = (bb ^ (bb & ERKKI_LOS_LIGHT)) | teho;
  }else if((bb & ERKKI_LOS_LIGHT) < teho) /* It's exposed but has less light than we're giving */
    bb = (bb ^ (bb & ERKKI_LOS_LIGHT)) | teho;
  /* And finally we update the seen area. */
  kentta->Kentta[k].data = bb;
}

/*
  Fastest possible LOF-routine the engine does have. We allocate the border locations of a
  rectangular area and use a simple way trying to reach it: at first we go in one of the
  main directions (up, down, right, left) until the border location is as far in X-direction
  as it is in Y-direction; then we start moving in both X- and Y-direction.

  This method leads to that a single LOS-beam is like a two-parted stick. Beam isn't direct
  and therefore not very realistic. And the area seen is rectangle-shaped. But it's fast!

*/
void ERKKI_see_area(KENTTA * kentta, signed int x, signed int y, signed int r, unsigned char teho){
signed int kpx = x - r, kpy = y - r, x1, y1, count, count2, q, k, t, rr = r * 8, a;

  count = 0; count2 = 0;
  /* There are rr (r*8) border locations */
  for(q=0;q<rr;q++){
    x1 = x; y1 = y;
    count2 = ( ABS(kpx-x1) > ABS(kpy - y1) ) ? 1:0; /* 1 if X-direction is more than Y-d. */
    t = teho;
    for(a = 0; a < r; a++){ /* We never move more than r steps */
      if(ERKKI_onkoeste(kentta, x1, y1, 0) == Este_Rajanyli)break;
      if(ABS(kpx-x1)==ABS(kpy-y1))count2 |= 2; /* Now we can go in both directions */
      ERKKI_spot_light(kentta, x1, y1, t); /* Lights on! */
      if(ERKKI_onkoeste(kentta, x1, y1, este_tyyli_fly | este_olio | este_tyyli_nako))break; /* este poks */
      if(t>0)t--; /* Light power goes down */
      if(count2 & 2){ /* We can go in both directions... */
        if(x1 < kpx)x1++; else if(x1 > kpx)x1--;
        if(y1 < kpy)y1++; else if(y1 > kpy)y1--;
      }else{ /* ... or not. */
        if(count2 & 1){ /* Move in X-dir towards goal. */
          if(x1 < kpx)x1++; else if(x1 > kpx)x1--;
        }else{ /* Or in Y-dir, same goal. */
          if(y1 < kpy)y1++; else if(y1 > kpy)y1--;
        }
      }
    }
    /* count shows, on what border we're moving. */
    if(count==0){kpx++; if(kpx >= x + r)count = 1;}
    else if(count==1){kpy++; if(kpy >= y + r)count = 2;}
    else if(count==2){kpx--; if(kpx <= x - r)count = 3;}
    else if(count==3){kpy--; if(kpy <= y - r)count = 4;}
  }
}

/*
  A clone method of see_area, but this time the area seen is corrected a bit to look more
  like a circle (we use sqrt). It isn't, but this method is still very fast!
*/
void ERKKI_see_area_corr(KENTTA * kentta, signed int x, signed int y, signed int r, unsigned char teho){
signed int kpx = x - r, kpy = y - r, x1, y1, count, count2, q, k, t, rr = r * 8, a;
double len;
  count = 0; count2 = 0;
  /* There are rr (r*8) border locations */
  for(q=0;q<rr;q++){
    x1 = abs(kpx-x); y1 = abs(kpy-y); /* Distance in axis between locations */
    count2 = ( x1 > y1 ) ? 1:0; /* 1 if X-direction is more than Y-d. */
    /* True distance to target location */
    len = sqrt(pow((double)x1, 2.0) + pow((double)y1, 2.0));
    /* How much we have to cut away? */
    len -= ((double)(len - r) * 1.75); /* This creates a beatiful circle :) */
    x1 = x; y1 = y;
    t = teho;
    /* Correction to number of steps */
    x1 = x; y1 = y;
    for(a = 0; a < (int)len; a++){ /* We never move more than r steps */
      if(ERKKI_onkoeste(kentta, x1, y1, 0) == Este_Rajanyli)break;
      if(ABS(kpx-x1)==ABS(kpy-y1))count2 |= 2; /* Now we can go in both directions */
      ERKKI_spot_light(kentta, x1, y1, t); /* Lights on! */
      if(ERKKI_onkoeste(kentta, x1, y1, este_tyyli_fly | este_olio | este_tyyli_nako))break; /* este poks */
      if(t>0)t--; /* Light power goes down */
      if(count2 & 2){ /* We can go in both directions... */
        if(x1 < kpx)x1++; else if(x1 > kpx)x1--;
        if(y1 < kpy)y1++; else if(y1 > kpy)y1--;
      }else{ /* ... or not. */
        if(count2 & 1){ /* Move in X-dir towards goal. */
          if(x1 < kpx)x1++; else if(x1 > kpx)x1--;
        }else{ /* Or in Y-dir, same goal. */
          if(y1 < kpy)y1++; else if(y1 > kpy)y1--;
        }
      }
    }
    /* count shows, on what border we're moving. */
    if(count==0){kpx++; if(kpx >= x + r)count = 1;}
    else if(count==1){kpy++; if(kpy >= y + r)count = 2;}
    else if(count==2){kpx--; if(kpx <= x - r)count = 3;}
    else if(count==3){kpy--; if(kpy <= y - r)count = 4;}
  }
}

/*
  This LOS-method goes through all borders of the field (kentta), using linear LOS-beam (calculated
  with simple method, count_angle_move - see erkki.c for information).
  We shoot a direct LOS-beam towards every border tile of the field.
  Becouse the method count_angle_move isn't trigonometrical, our field seen would be (once more)
  tetragonal. I calculated my very own way to correct the number of steps taken from (x,y) to
  make the area look more realistic, using the terms xm_ and ym_ (movement in X- and Y-directions
  to make a direct line - another is always 1 and the second one < 1). Theory is explained in
  erkki.c in the code of cout_angle_move.
*/
void ERKKI_see_area_real(KENTTA * kentta, signed int x, signed int y, signed int r, unsigned char teho){
double xm_, ym_, x1, y1, eta;
signed int kpx = x-r, kpy = y-r, count = 0, m, q, t, num = 2*(kentta->w + kentta->h-2), k;
signed int xxx, yyy;
  for(q = 0; q < num; q++){
    /* Steps between goal location and start location */
    count_angle_move((double)(x), (double)(y), (double)(kpx), (double)(kpy), &xm_, &ym_);
    x1 = d_abs((double)(x-kpx)); y1 = d_abs((double)(y-kpy));
    /* True distance between locations */
    eta = sqrt(pow(x1, 2.0) + pow(y1, 2.0));
    /* How much we have to cut away? Oikea askelmr (t) on pienin etisyys reunaan :/ */
    eta -= ((double)(eta - r) * 1.75); /* This creates a beatiful circle :) */
    x1 = (double)x; y1 = (double)y; t = teho;
    /* And then we go. matka is the 'realistic' number of steps. */
    for(m = 0; m < (int)eta; m++){
      /* Rounding the locations */
      xxx = (int)(x1+0.5); yyy = (int)(y1+0.5);
      /* Are we out of field? */
      if(ERKKI_onkoeste(kentta, xxx, yyy, 0) == Este_Rajanyli)break;

      ERKKI_spot_light(kentta, xxx, yyy, t); /* Lights on! */

      /* Facing a wall or another not-see-trough-object? */
      if(ERKKI_onkoeste(kentta, xxx, yyy, este_tyyli_fly | este_olio | este_tyyli_nako))break;
      if(t>0)t--; /* Light power goes down */
      /* Move to next tile */
      x1 += xm_; y1 += ym_;
    }
    /* Round, round, round we go... */
    if(count==0){kpx++; if(kpx >= x + r)count = 1;}
    else if(count==1){kpy++; if(kpy >= y + r)count = 2;}
    else if(count==2){kpx--; if(kpx <= x - r)count = 3;}
    else if(count==3){kpy--; if(kpy <= y - r)count = 4;}
  }
}

/*
  Jus like below, but we're using trigonometry to calculate movements instead of count_angle_move.
  Also, we don't have to make any corrections to area seen - it's perfectly round.

  To make area to look good, we cannot just project LOS-beam to border tiles; what if the start location
  is just on the border? Then the beams goes a bit wrong and the result looks terrible (I've tried).
  So, we move the whole field so that start location is allways in the middle of the field when we
  calculate the movement and after that we place it to the right location.
*/
void ERKKI_see_area_trig(KENTTA * kentta, signed int x, signed int y, signed int r, unsigned char teho){
double angle, x1, y1, xm_, ym_;
signed int kpx = 0, kpy = 0, count = 0, m, q, t, num = 2*(kentta->w + kentta->h-2), k, xx, yy, mx, my;
  /* Position in relation to middle point of the field */
  mx = x - (int)((kentta->w+1)/2); my = y - (int)((kentta->h+1)/2);
  for(q = 0; q < num; q++){
    x1 = (double)(kentta->w)/2.0; y1 = (double)(kentta->h)/2.0; /* Jump to the middle */
    if(x1 == kpx)angle = (M_PI / 2.0) * ((kpy <= y1)?(1):(-1)); /* Special point */
    else{
      /* As we remember, tan(angle) = Y/X -> angle = atan(Y/X) */
      angle = atan( (double)(kpy-y1) / (double)(x1-kpx) );
      if(count > 1)angle = M_PI - angle; /* Mirror the angle */
    }
    t = teho;
    xm_ = cos(angle); ym_ = sin(angle);
    /* On we go */
    for(m = 0; m < r; m++){
      /* In here we move back to real position - and round the coords. */
      xx = mx + (int)(x1+0.5); yy = my + (int)(y1+0.5);
      /* Out? */
      if(ERKKI_onkoeste(kentta, xx, yy, 0) == Este_Rajanyli)break;
      ERKKI_spot_light(kentta, xx, yy, t); /* Lights on! */
      /* Blocked? */
      if(ERKKI_onkoeste(kentta, xx, yy, este_tyyli_fly | este_olio | este_tyyli_nako))break;
      if(t>0)t--; /* Lights goes down */
      /* And move the beam */
      x1 += xm_; y1 += ym_;
    }
    /* Going around the field */
    if(count==0){kpx++; if(kpx == kentta->w - 1)count = 1;}
    else if(count==1){kpy++; if(kpy >= kentta->h)count = 2;}
    else if(count==2){kpx--; if(kpx <= 0)count = 3;}
    else if(count==3){kpy--; if(kpy <= 0)count = 4;}
  }
}

/*
  Is there a LOS from (x1,y1) to (x2,y2)? This method uses two-lines method (not so exact but fast).
*/
int InLineOfSight(KENTTA * kentta, signed int x1, signed int y1, signed int x2, signed int y2){
signed int x,y,rr,a;
unsigned char r=0;
  if( NOTINMAP(kentta, x1, y1) || NOTINMAP(kentta, x2, y2) )return ERKKI_ERROR;
  x = x1; y = y1;
  if( ABS(x2-x1) > ABS(y2-y1) ){ /* Which dir. we go first? */
    r |= 2;
    rr = ABS(x2-x1); /* Steps between locations */
  }else{
    rr = ABS(y2-y1); /* Steps between locations */
  }
  for(a=0;a<rr;a++){
    if( ABS(x-x2) == ABS(y-y2) )r |= 1; /* Start going in both directions */
    /* If way's blocked, there's no LOS. */
    if(ERKKI_onkoeste(kentta, x, y, este_tyyli_fly | este_olio | este_tyyli_nako))return ERKKI_NOFOUND;
    if(r & 1){ /* Can go both dirs */
      if(x < x2)x++; else if(x > x2)x--;
      if(y < y2)y++; else if(y > y2)y--;
    }else{
      if(r & 2){
        if(x < x2)x++; else if(x > x2)x--;
      }else{
        if(y < y2)y++; else if(y > y2)y--;
      }
    }
  }
  return ERKKI_OK; /* There is LOS from (x1, y1) to (x2, y2) */
}

/*
  Like the one before, but in this movement is calculated with count_angle_move.
*/
int InLineOfSight_real(KENTTA * kentta, signed int x1, signed int y1, signed int x2, signed int y2){
signed int rr,a;
unsigned char r=0;
double xm, ym, x, y;
  if( NOTINMAP(kentta, x1, y1) || NOTINMAP(kentta, x2, y2) )return ERKKI_ERROR;
  x = (double)x1; y = (double)y1;
  count_angle_move((double)x1, (double)y1, (double)x2, (double)y2, &xm, &ym);
  if( ABS(x2-x1) > ABS(y2-y1) ){
    rr = (int)ABS(x2-x1); /* Steps between locations */
  }else{
    rr = (int)ABS(y2-y1); /* Steps between locations */
  }
  for(a=0;a<rr;a++){
    /* If way's blocked, there's no LOS. */
    if(ERKKI_onkoeste(kentta, x, y, este_tyyli_fly | este_olio | este_tyyli_nako))return ERKKI_NOFOUND;
    x += xm; y += ym;
  }
  return ERKKI_OK; /* There is LOS from (x1, y1) to (x2, y2) */
}

